from typing import Any, Dict

from pylabrobot.serializer import SerializableMixin
from pylabrobot.utils.interpolation import interpolate_1d


class HamiltonLiquidClass(SerializableMixin):
  """A liquid class like used in VENUS / Venus on Vantage."""

  def __init__(
    self,
    curve: Dict[float, float],
    aspiration_flow_rate: float,
    aspiration_mix_flow_rate: float,
    aspiration_air_transport_volume: float,
    aspiration_blow_out_volume: float,
    aspiration_swap_speed: float,
    aspiration_settling_time: float,
    aspiration_over_aspirate_volume: float,
    aspiration_clot_retract_height: float,
    dispense_flow_rate: float,
    dispense_mode: float,
    dispense_mix_flow_rate: float,
    dispense_air_transport_volume: float,
    dispense_blow_out_volume: float,
    dispense_swap_speed: float,
    dispense_settling_time: float,
    dispense_stop_flow_rate: float,
    dispense_stop_back_volume: float,
  ):
    self.curve = curve

    self.aspiration_flow_rate = aspiration_flow_rate
    self.aspiration_mix_flow_rate = aspiration_mix_flow_rate
    self.aspiration_air_transport_volume = aspiration_air_transport_volume
    self.aspiration_blow_out_volume = aspiration_blow_out_volume
    self.aspiration_swap_speed = aspiration_swap_speed
    self.aspiration_settling_time = aspiration_settling_time
    self.aspiration_over_aspirate_volume = aspiration_over_aspirate_volume
    self.aspiration_clot_retract_height = aspiration_clot_retract_height

    self.dispense_mode = dispense_mode
    self.dispense_flow_rate = dispense_flow_rate
    self.dispense_mix_flow_rate = dispense_mix_flow_rate
    self.dispense_air_transport_volume = dispense_air_transport_volume
    self.dispense_blow_out_volume = dispense_blow_out_volume
    self.dispense_swap_speed = dispense_swap_speed
    self.dispense_settling_time = dispense_settling_time
    self.dispense_stop_flow_rate = dispense_stop_flow_rate
    self.dispense_stop_back_volume = dispense_stop_back_volume

  def compute_corrected_volume(self, target_volume: float) -> float:
    """Compute the piston displacement volume needed to achieve a desired liquid volume.

    This method determines how far the pipette piston must move (i.e., the commanded
    internal volume) to aspirate or dispense a specified *target liquid volume*.
    Because factors such as air compressibility, liquid viscosity, and tip geometry
    affect the relationship between piston displacement and actual transferred volume,
    Hamilton liquid classes use an empirically derived **correction curve**.

    The correction curve maps *nominal liquid volumes* (target) to the corresponding
    *piston displacement volumes* required to achieve them. If the requested
    `target_volume` exactly matches a calibration point, its mapped value is used.
    Otherwise, the function performs **piecewise linear interpolation** between the
    two nearest calibration points. For values outside the calibration range,
    the nearest segment is linearly extrapolated.

    This interpolation is performed using
    :func:`pylabrobot.utils.interpolation.interpolate_1d`.

    Args:
      target_volume: The liquid volume to be aspirated or dispensed (in µL).

    Returns:
      The corrected piston displacement volume (in µL) that the pipette mechanism
      must execute to achieve the desired liquid transfer.

    Raises:
      ValueError: If the correction curve data is invalid or non-numeric.
    """
    if self.curve is None:
      return target_volume

    return interpolate_1d(target_volume, self.curve, bounds_handling="extrapolate")

  def serialize(self) -> Dict[str, Any]:
    """Serialize the liquid class to a dictionary."""
    return {
      "curve": self.curve,
      "aspiration_flow_rate": self.aspiration_flow_rate,
      "aspiration_mix_flow_rate": self.aspiration_mix_flow_rate,
      "aspiration_air_transport_volume": self.aspiration_air_transport_volume,
      "aspiration_blow_out_volume": self.aspiration_blow_out_volume,
      "aspiration_swap_speed": self.aspiration_swap_speed,
      "aspiration_settling_time": self.aspiration_settling_time,
      "aspiration_over_aspirate_volume": self.aspiration_over_aspirate_volume,
      "aspiration_clot_retract_height": self.aspiration_clot_retract_height,
      "dispense_mode": self.dispense_mode,
      "dispense_flow_rate": self.dispense_flow_rate,
      "dispense_mix_flow_rate": self.dispense_mix_flow_rate,
      "dispense_air_transport_volume": self.dispense_air_transport_volume,
      "dispense_blow_out_volume": self.dispense_blow_out_volume,
      "dispense_swap_speed": self.dispense_swap_speed,
      "dispense_settling_time": self.dispense_settling_time,
      "dispense_stop_flow_rate": self.dispense_stop_flow_rate,
      "dispense_stop_back_volume": self.dispense_stop_back_volume,
    }



"""Corning-Falcon Tubes"""

from pylabrobot.resources.tube import Tube

# # # # # # # # # # Cor_Falcon_tube_50mL_Vb # # # # # # # # # #


def Cor_Falcon_tube_50mL_Vb(name: str) -> Tube:
  """
  Corning cat. no.: 352098
  - manufacturer_link: https://ecatalog.corning.com/life-sciences/b2b/UK/en/Liquid-Handling/Tubes%2C-Liquid-Handling/Centrifuge-Tubes/Falcon%C2%AE-Conical-Centrifuge-Tubes/p/falconConicalTubes
  - distributor: (Fisher Scientific, 14-959-49A)
  - brand: Falcon
  - material: Polypropylene
  - tech_drawing: tech_drawings/Cor_Falcon_tube_50mL.pdf
  - cap_style: screw-cap
  """

  diameter = 30
  return Tube(
    name=name,
    size_x=diameter,
    size_y=diameter,
    size_z=115,
    model="Falcon 50mL",
    max_volume=50_000,
    material_z_thickness=1.2,
  )


# # # # # # # # # # Cor_Falcon_tube_15mL_Vb # # # # # # # # # #


def Cor_Falcon_tube_15mL_Vb(name: str) -> Tube:
  """
  Corning cat. no.: 352196
  - manufacturer_link: https://ecatalog.corning.com/life-sciences/b2b/UK/en/Liquid-Handling/Tubes%2C-Liquid-Handling/Centrifuge-Tubes/Falcon%C2%AE-Conical-Centrifuge-Tubes/p/falconConicalTubes
  - distributor: (Fisher Scientific, 14-959-53A)
  - brand: Falcon
  - material: Polypropylene
  - tech_drawing: tech_drawings/Cor_Falcon_tube_15mL_Vb.pdf
  - cap_style: screw-cap
  """

  diameter = 17
  return Tube(
    name=name,
    size_x=diameter,
    size_y=diameter,
    size_z=120,
    model="Falcon 15mL",
    max_volume=15_000,
  )


# # # # # # # # # # Cor_Falcon_tube_14mL_Rb # # # # # # # # # #


def Cor_Falcon_tube_14mL_Rb(name: str) -> Tube:
  """
  Corning cat. no.: 352059
  - manufacturer_link: https://ecatalog.corning.com/life-sciences/b2b/UK/en/General-Labware/Tubes/Tubes%2C-Round-Bottom/Falcon%C2%AE-Round-Bottom-High-clarity-Polypropylene-Tube/p/highClarityPolypropyleneRoundBottomTubes
  - distributor: (Fisher Scientific, 10110101)
  - brand: Falcon
  - material: Polypropylene
  - tech_drawing: tech_drawings/Cor_Falcon_tube_14mL_Rb.pdf
  - cap_style: snap-cap
  """

  diameter = 17
  return Tube(
    name=name,
    size_x=diameter,
    size_y=diameter,
    size_z=95,
    model="Falcon_tube_14mL_Rb",
    material_z_thickness=1.19,
    max_volume=14_000,  # units: ul
  )


# Previous names in PLR:


def falcon_tube_50mL(name: str) -> Tube:
  raise NotImplementedError(
    "falcon_tube_50mL definition is deprecated. Use Cor_Falcon_tube_50mL instead."
  )


def falcon_tube_15mL(name: str) -> Tube:
  raise NotImplementedError(
    "falcon_tube_15mL definition is deprecated. Use Cor_Falcon_tube_15mL_Vb instead."
  )


def Falcon_tube_14mL_Rb(name: str) -> Tube:
  raise NotImplementedError(
    "Falcon_tube_14mL_Rb definition is deprecated. Use Cor_Falcon_tube_14mL_Rb instead."
  )



import asyncio
import sys
import unittest
from unittest.mock import AsyncMock, MagicMock, patch

from pylabrobot.plate_reading.tecan.spark20m.enums import SparkDevice
from pylabrobot.plate_reading.tecan.spark20m.spark_backend import ExperimentalSparkBackend
from pylabrobot.resources.plate import Plate
from pylabrobot.resources.well import Well

sys.modules["usb.core"] = MagicMock()
sys.modules["usb.util"] = MagicMock()


class TestExperimentalSparkBackend(unittest.IsolatedAsyncioTestCase):
  async def asyncSetUp(self) -> None:
    # Patch SparkReaderAsync
    self.reader_patcher = patch(
      "pylabrobot.plate_reading.tecan.spark20m.spark_backend.SparkReaderAsync"
    )
    self.MockReaderClass = self.reader_patcher.start()
    self.mock_reader = self.MockReaderClass.return_value

    self.mock_reader.connect = AsyncMock()
    self.mock_reader.close = AsyncMock()
    self.mock_reader.send_command = AsyncMock(return_value=True)

    # Patch processor functions
    self.abs_proc_patcher = patch(
      "pylabrobot.plate_reading.tecan.spark20m.spark_backend.process_absorbance"
    )
    self.mock_process_absorbance = self.abs_proc_patcher.start()

    self.fluo_proc_patcher = patch(
      "pylabrobot.plate_reading.tecan.spark20m.spark_backend.process_fluorescence"
    )
    self.mock_process_fluorescence = self.fluo_proc_patcher.start()

    self.backend = ExperimentalSparkBackend()
    # Populate devices so device-connected checks pass
    self.backend.reader.devices = {
      SparkDevice.ABSORPTION: MagicMock(),
      SparkDevice.FLUORESCENCE: MagicMock(),
      SparkDevice.PLATE_TRANSPORT: MagicMock(),
    }

  async def asyncTearDown(self) -> None:
    self.reader_patcher.stop()
    self.abs_proc_patcher.stop()
    self.fluo_proc_patcher.stop()

  async def test_setup(self) -> None:
    await self.backend.setup()
    self.mock_reader.connect.assert_called_once()
    # Verify that send_command was called for init_module
    self.mock_reader.send_command.assert_called()

  async def test_open(self) -> None:
    await self.backend.open()
    self.mock_reader.send_command.assert_called()

  async def test_read_absorbance(self) -> None:
    # Mock background read
    stop_event = MagicMock()
    bg_task: "asyncio.Future[None]" = asyncio.Future()
    bg_task.set_result(None)
    self.mock_reader.start_background_read = AsyncMock(return_value=(bg_task, stop_event, []))

    self.mock_process_absorbance.return_value = [[0.5]]

    plate = MagicMock(spec=Plate)
    plate.num_items_x = 2
    plate.num_items_y = 2
    plate.get_size_y.return_value = 100

    well = MagicMock(spec=Well)
    well.parent = plate
    well.get_row.return_value = 0
    well.get_column.return_value = 0
    well.location = MagicMock()
    well.location.x = 0
    well.location.y = 0
    well.location.z = 0
    well.get_anchor.return_value = MagicMock()

    plate.get_item.return_value = well

    results = await self.backend.read_absorbance(plate, None, wavelength=600)

    self.assertEqual(len(results), 1)
    self.assertEqual(results[0]["wavelength"], 600)
    self.assertEqual(results[0]["data"], [[0.5]])

  async def test_read_fluorescence(self) -> None:
    # Mock background read
    stop_event = MagicMock()
    bg_task: "asyncio.Future[None]" = asyncio.Future()
    bg_task.set_result(None)
    self.mock_reader.start_background_read = AsyncMock(return_value=(bg_task, stop_event, []))

    self.mock_process_fluorescence.return_value = [[100.0]]

    plate = MagicMock(spec=Plate)
    plate.num_items_x = 2
    plate.num_items_y = 2
    plate.get_size_y.return_value = 100

    well = MagicMock(spec=Well)
    well.parent = plate
    well.get_row.return_value = 0
    well.get_column.return_value = 0
    well.location = MagicMock()
    well.location.x = 0
    well.location.y = 0
    well.location.z = 0
    well.get_anchor.return_value = MagicMock()

    plate.get_item.return_value = well

    results = await self.backend.read_fluorescence(
      plate, [well], excitation_wavelength=480, emission_wavelength=520
    )

    self.assertEqual(len(results), 1)
    self.assertEqual(results[0]["ex_wavelength"], 480)
    self.assertEqual(results[0]["em_wavelength"], 520)
    self.assertEqual(results[0]["data"], [[100.0]])

  async def test_get_average_temperature(self) -> None:
    # Mock reader messages
    self.mock_reader.msgs = [
      {"number": 100, "args": ["2500"]},  # 25.00
      {"number": 100, "args": ["2600"]},  # 26.00
      {"number": 200, "args": ["something else"]},
    ]

    temp = await self.backend.get_average_temperature()
    self.assertEqual(temp, 25.5)

  async def test_get_average_temperature_empty(self) -> None:
    self.mock_reader.msgs = []
    temp = await self.backend.get_average_temperature()
    self.assertIsNone(temp)


if __name__ == "__main__":
  unittest.main()



import unittest

from pylabrobot.centrifuge import (
  BucketHasPlateError,
  BucketNoPlateError,
  Centrifuge,
  CentrifugeDoorError,
  Loader,
  LoaderNoPlateError,
  NotAtBucketError,
)
from pylabrobot.centrifuge.backend import CentrifugeBackend, LoaderBackend
from pylabrobot.centrifuge.chatterbox import CentrifugeChatterboxBackend, LoaderChatterboxBackend
from pylabrobot.resources import Coordinate, Cor_96_wellplate_360ul_Fb


class CentrifugeTests(unittest.IsolatedAsyncioTestCase):
  def test_serialization(self):
    centrifuge = Centrifuge(
      backend=CentrifugeChatterboxBackend(),
      name="centrifuge",
      size_x=1,
      size_y=1,
      size_z=1,
    )
    serialized = centrifuge.serialize()
    deserialized = Centrifuge.deserialize(serialized)
    self.assertEqual(deserialized, centrifuge)


class CentrifugeLoaderResourceModelTests(unittest.IsolatedAsyncioTestCase):
  async def asyncSetUp(self):
    self.mock_centrifuge_backend = unittest.mock.MagicMock(spec=CentrifugeBackend)
    self.mock_loader_backend = unittest.mock.MagicMock(spec=LoaderBackend)
    self.centrifuge = Centrifuge(
      backend=self.mock_centrifuge_backend, name="centrifuge", size_x=1, size_y=1, size_z=1
    )
    self.loader = Loader(
      backend=self.mock_loader_backend,
      centrifuge=self.centrifuge,
      name="loader",
      size_x=1,
      size_y=1,
      size_z=1,
      child_location=Coordinate.zero(),
    )
    self.plate = Cor_96_wellplate_360ul_Fb(name="plate")
    return await super().asyncSetUp()

  async def test_go_to_bucket(self):
    self.assertIsNone(self.centrifuge.at_bucket)
    await self.centrifuge.go_to_bucket1()
    self.assertEqual(self.centrifuge.at_bucket, self.centrifuge.bucket1)
    await self.centrifuge.go_to_bucket2()
    self.assertEqual(self.centrifuge.at_bucket, self.centrifuge.bucket2)

  async def test_load(self):
    await self.centrifuge.go_to_bucket1()
    await self.centrifuge.open_door()
    assert self.centrifuge._door_open
    assert self.centrifuge.door_open
    self.loader.assign_child_resource(self.plate)
    await self.loader.load()
    self.mock_loader_backend.load.assert_awaited_once()
    assert self.centrifuge.at_bucket is not None
    self.assertEqual(self.centrifuge.at_bucket.children[0], self.plate)
    self.assertEqual(self.loader.children, [])

  async def test_load_locked_door(self):
    self.loader.assign_child_resource(self.plate)
    with self.assertRaises(CentrifugeDoorError):
      await self.loader.load()
    self.mock_loader_backend.load.assert_not_awaited()

  async def test_load_no_plate(self):
    await self.centrifuge.go_to_bucket1()
    await self.centrifuge.open_door()
    with self.assertRaises(LoaderNoPlateError):
      await self.loader.load()
    self.mock_loader_backend.load.assert_not_awaited()

  async def test_load_bucket_has_plate(self):
    await self.centrifuge.go_to_bucket1()
    await self.centrifuge.open_door()
    assert self.centrifuge.at_bucket is not None
    self.centrifuge.at_bucket.assign_child_resource(self.plate)
    another_plate = Cor_96_wellplate_360ul_Fb(name="another_plate")
    self.loader.assign_child_resource(another_plate)
    with self.assertRaises(BucketHasPlateError):
      await self.loader.load()
    self.mock_loader_backend.load.assert_not_awaited()

  async def test_load_not_at_bucket(self):
    self.loader.assign_child_resource(self.plate)
    await self.centrifuge.open_door()
    with self.assertRaises(NotAtBucketError):
      await self.loader.load()
    self.mock_loader_backend.load.assert_not_awaited()

  async def test_unload(self):
    await self.centrifuge.go_to_bucket1()
    await self.centrifuge.open_door()
    assert self.centrifuge.at_bucket is not None
    self.centrifuge.at_bucket.assign_child_resource(self.plate)
    await self.loader.unload()
    self.mock_loader_backend.unload.assert_awaited_once()
    self.assertEqual(self.centrifuge.at_bucket.children, [])
    self.assertEqual(self.loader.children, [self.plate])

  async def test_unload_locked_door(self):
    self.loader.assign_child_resource(self.plate)
    with self.assertRaises(CentrifugeDoorError):
      await self.loader.unload()
    self.mock_loader_backend.unload.assert_not_awaited()

  async def test_unload_bucket_has_no_plate(self):
    await self.centrifuge.go_to_bucket1()
    await self.centrifuge.open_door()
    with self.assertRaises(BucketNoPlateError):
      await self.loader.unload()
    self.mock_loader_backend.unload.assert_not_awaited()

  async def test_unload_loader_has_plate(self):
    await self.centrifuge.go_to_bucket1()
    await self.centrifuge.open_door()
    self.loader.assign_child_resource(self.plate)
    with self.assertRaises(BucketNoPlateError):
      await self.loader.unload()
    self.mock_loader_backend.unload.assert_not_awaited()

  async def test_unload_not_at_bucket(self):
    self.loader.assign_child_resource(self.plate)
    await self.centrifuge.open_door()
    with self.assertRaises(NotAtBucketError):
      await self.loader.unload()
    self.mock_loader_backend.unload.assert_not_awaited()

  def test_serialize(self):
    self.loader.backend = LoaderChatterboxBackend()
    self.centrifuge.backend = CentrifugeChatterboxBackend()
    serialized = self.loader.serialize()
    self.assertEqual(Loader.deserialize(serialized), self.loader)



import unittest
from unittest.mock import patch

import pytest

pytest.importorskip("ot_api")

from pylabrobot.resources.itemized_resource import ItemizedResource
from pylabrobot.thermocycling.opentrons import OpentronsThermocyclerModuleV1
from pylabrobot.thermocycling.opentrons_backend import OpentronsThermocyclerBackend
from pylabrobot.thermocycling.standard import BlockStatus, LidStatus, Protocol, Stage, Step


class TestOpentronsThermocyclerBackend(unittest.IsolatedAsyncioTestCase):
  async def asyncSetUp(self):
    await super().asyncSetUp()
    self.thermocycler_backend = OpentronsThermocyclerBackend(opentrons_id="test_id")

  def test_opentrons_v1_serialization(self):
    """Test that the Opentrons-specific resource model serializes correctly."""
    tc_model = OpentronsThermocyclerModuleV1(
      name="test_v1_tc",
      opentrons_id="test_id",
      child=ItemizedResource(name="plate", size_x=1, size_y=1, size_z=1, ordered_items={}),
    )
    serialized = tc_model.serialize()
    assert "opentrons_id" in serialized
    assert serialized["opentrons_id"] == "test_id"
    deserialized = OpentronsThermocyclerModuleV1.deserialize(serialized)
    assert tc_model == deserialized

  @patch("pylabrobot.thermocycling.opentrons_backend.list_connected_modules")
  async def test_find_module_raises_error_if_not_found(self, mock_list_connected_modules):
    """Test that an error is raised if the module is not found."""
    mock_list_connected_modules.return_value = [{"id": "some_other_id", "data": {}}]
    with self.assertRaises(RuntimeError) as e:
      await self.thermocycler_backend.get_lid_open()
    self.assertEqual(str(e.exception), "Module 'test_id' not found")

  @patch("pylabrobot.thermocycling.opentrons_backend.thermocycler_open_lid")
  async def test_open_lid(self, mock_open_lid):
    await self.thermocycler_backend.open_lid()
    mock_open_lid.assert_called_once_with(module_id="test_id")

  @patch("pylabrobot.thermocycling.opentrons_backend.thermocycler_close_lid")
  async def test_close_lid(self, mock_close_lid):
    await self.thermocycler_backend.close_lid()
    mock_close_lid.assert_called_once_with(module_id="test_id")

  @patch("pylabrobot.thermocycling.opentrons_backend.thermocycler_set_block_temperature")
  async def test_set_block_temperature(self, mock_set_block_temp):
    await self.thermocycler_backend.set_block_temperature([95.0])
    mock_set_block_temp.assert_called_once_with(celsius=95.0, module_id="test_id")

  @patch("pylabrobot.thermocycling.opentrons_backend.thermocycler_set_lid_temperature")
  async def test_set_lid_temperature(self, mock_set_lid_temp):
    await self.thermocycler_backend.set_lid_temperature([105.0])
    mock_set_lid_temp.assert_called_once_with(celsius=105.0, module_id="test_id")

  @patch("pylabrobot.thermocycling.opentrons_backend.thermocycler_deactivate_block")
  async def test_deactivate_block(self, mock_deactivate_block):
    await self.thermocycler_backend.deactivate_block()
    mock_deactivate_block.assert_called_once_with(module_id="test_id")

  @patch("pylabrobot.thermocycling.opentrons_backend.thermocycler_deactivate_lid")
  async def test_deactivate_lid(self, mock_deactivate_lid):
    await self.thermocycler_backend.deactivate_lid()
    mock_deactivate_lid.assert_called_once_with(module_id="test_id")

  @patch("pylabrobot.thermocycling.opentrons_backend.thermocycler_run_profile_no_wait")
  async def test_run_protocol(self, mock_run_profile):
    protocol = Protocol(stages=[Stage(steps=[Step(temperature=[95], hold_seconds=10)], repeats=1)])
    await self.thermocycler_backend.run_protocol(protocol, 50.0)
    # print all calls
    mock_run_profile.assert_called_once_with(
      profile=[{"celsius": 95, "holdSeconds": 10}], block_max_volume=50.0, module_id="test_id"
    )

  @patch("pylabrobot.thermocycling.opentrons_backend.list_connected_modules")
  async def test_getters_return_correct_data(self, mock_list_connected_modules):
    mock_data = {
      "id": "test_id",
      "data": {
        "currentTemperature": 25.5,
        "targetTemperature": 95.0,
        "lidTemperature": 37.1,
        "lidTargetTemperature": 105.0,
        "lidStatus": "open",
        "lidTemperatureStatus": "holding at target",
        "status": "holding at target",
        "holdTime": 12.0,
        "currentCycleIndex": 2,
        "totalCycleCount": 10,
        "currentStepIndex": 1,
        "totalStepCount": 3,
      },
    }
    mock_list_connected_modules.return_value = [mock_data]

    assert await self.thermocycler_backend.get_block_current_temperature() == [25.5]
    assert await self.thermocycler_backend.get_block_target_temperature() == [95.0]
    assert await self.thermocycler_backend.get_lid_current_temperature() == [37.1]
    assert await self.thermocycler_backend.get_lid_target_temperature() == [105.0]
    assert await self.thermocycler_backend.get_lid_open() is True
    assert await self.thermocycler_backend.get_lid_status() == LidStatus.HOLDING_AT_TARGET
    assert await self.thermocycler_backend.get_block_status() == BlockStatus.HOLDING_AT_TARGET
    assert await self.thermocycler_backend.get_hold_time() == 12.0
    # assert await self.thermocycler_backend.get_current_cycle_index() == 1  # 2 - 1 = 1 (zero-based)
    # assert await self.thermocycler_backend.get_total_cycle_count() == 10
    assert await self.thermocycler_backend.get_current_step_index() == 0  # 1 - 1 = 0 (zero-based)
    assert await self.thermocycler_backend.get_total_step_count() == 3



from __future__ import annotations

from typing import Any

from .agilent import (
  BioTekPlateReaderBackend,
  Cytation5Backend,
  Cytation5ImagingConfig,
  CytationBackend,
  CytationImagingConfig,
  SynergyH1Backend,
)
from .bmg_labtech import CLARIOstarBackend
from .byonoy import (
  ByonoyAbsorbance96AutomateBackend,
  ByonoyLuminescence96AutomateBackend,
)
from .chatterbox import PlateReaderChatterboxBackend
from .image_reader import ImageReader
from .imager import Imager
from .molecular_devices import (
  Calibrate,
  CarriageSpeed,
  KineticSettings,
  MolecularDevicesBackend,
  MolecularDevicesError,
  MolecularDevicesFirmwareError,
  MolecularDevicesHardwareError,
  MolecularDevicesMotionError,
  MolecularDevicesNVRAMError,
  MolecularDevicesSettings,
  MolecularDevicesSpectraMax384PlusBackend,
  MolecularDevicesSpectraMaxM5Backend,
  MolecularDevicesUnrecognizedCommandError,
  PmtGain,
  ReadMode,
  ReadOrder,
  ReadType,
  ShakeSettings,
  SpectrumSettings,
)
from .plate_reader import PlateReader
from .standard import (
  Exposure,
  FocalPosition,
  Gain,
  ImagingMode,
  ImagingResult,
  Objective,
)
from .tecan import ExperimentalTecanInfinite200ProBackend
from .tecan.spark20m.spark_backend import ExperimentalSparkBackend



from typing import List, Optional, TypeVar

T = TypeVar("T")


def fill_in_defaults(val: Optional[List[T]], default: List[T]) -> List[T]:
  """Util for converting an argument to the appropriate format for low level star methods."""
  # if the val is None, use the default.
  if val is None:
    return default
  # if the val is a list, it must be of the correct length.
  if len(val) != len(default):
    raise ValueError(f"Value length must equal num operations ({len(default)}), but is {val}")
  # replace None values in list with default values.
  val = [v if v is not None else d for v, d in zip(val, default)]
  # the value is ready to be used.
  return val



import math
import unittest

from pylabrobot.resources.height_volume_functions import (
  calculate_liquid_volume_container_2segments_round_ubottom,
  calculate_liquid_volume_container_2segments_round_vbottom,
  calculate_liquid_volume_container_2segments_square_ubottom,
  calculate_liquid_volume_container_2segments_square_vbottom,
)


class TestVolumeFunctions(unittest.TestCase):
  """Tests for the volume functions"""

  def test_calculate_liquid_volume_container_2segments_square_vbottom(
    self,
  ):
    # Exactly the full pyramid
    self.assertAlmostEqual(
      calculate_liquid_volume_container_2segments_square_vbottom(
        x=10, y=10, h_pyramid=10, h_cube=10, liquid_height=10
      ),
      (10 * 10 * 10) / 3,
    )
    # Test for liquid height within the pyramid

    self.assertAlmostEqual(
      calculate_liquid_volume_container_2segments_square_vbottom(
        x=10, y=10, h_pyramid=10, h_cube=10, liquid_height=5
      ),
      (10 * 10 * 10) / 3 * (5 / 10) ** 3,
    )

    # Test for liquid height extending into the cube
    self.assertAlmostEqual(
      calculate_liquid_volume_container_2segments_square_vbottom(
        x=10, y=10, h_pyramid=10, h_cube=10, liquid_height=15
      ),
      (10 * 10 * 10) / 3 + 10 * 10 * 5,
    )

  def test_calculate_liquid_volume_container_2segments_round_vbottom(
    self,
  ):
    # Exactly the full cone
    self.assertAlmostEqual(
      calculate_liquid_volume_container_2segments_round_vbottom(
        d=10, h_cone=10, h_cylinder=10, liquid_height=10
      ),
      math.pi * (10 / 2) ** 2 * 10 / 3,
    )

    # Test for liquid height within the cone
    self.assertAlmostEqual(
      calculate_liquid_volume_container_2segments_round_vbottom(
        d=10, h_cone=10, h_cylinder=10, liquid_height=5
      ),
      math.pi * (10 / 2) ** 2 * 10 / 3 * (5 / 10) ** 3,
    )

    # Test for liquid height extending into the cylinder
    self.assertAlmostEqual(
      calculate_liquid_volume_container_2segments_round_vbottom(
        d=10, h_cone=10, h_cylinder=10, liquid_height=15
      ),
      math.pi * (10 / 2) ** 2 * 10 / 3 + math.pi * (10 / 2) ** 2 * 5,
    )

  def test_calculate_liquid_volume_container_2segments_round_ubottom(
    self,
  ):
    # Exactly half the sphere
    self.assertAlmostEqual(
      calculate_liquid_volume_container_2segments_round_ubottom(
        d=10, h_cylinder=10, liquid_height=5
      ),
      (2 / 3) * math.pi * 5**3,
    )

    # Test for liquid height within the half sphere
    self.assertAlmostEqual(
      calculate_liquid_volume_container_2segments_round_ubottom(
        d=10, h_cylinder=10, liquid_height=5
      ),
      (1 / 3) * math.pi * 5**2 * (3 * 5 - 5),
    )

    # Test for liquid height extending into the cylinder
    self.assertAlmostEqual(
      calculate_liquid_volume_container_2segments_round_ubottom(
        d=10, h_cylinder=10, liquid_height=15
      ),
      (2 / 3) * math.pi * 5**3 + math.pi * (10 / 2) ** 2 * 10,
    )

  def test_calculate_liquid_volume_container_2segments_square_ubottom(
    self,
  ):
    # Exactly the full hemisphere
    self.assertAlmostEqual(
      calculate_liquid_volume_container_2segments_square_ubottom(
        x=10, h_cuboid=10, liquid_height=5
      ),
      (2 / 3) * math.pi * (10 / 2) ** 3,
    )

    # Test for liquid height within the hemisphere
    self.assertAlmostEqual(
      calculate_liquid_volume_container_2segments_square_ubottom(
        x=10, h_cuboid=10, liquid_height=5
      ),
      (1 / 3) * math.pi * 5**2 * (3 * 5 - 5),
    )

    # Test for liquid height extending into the cuboid
    self.assertAlmostEqual(
      calculate_liquid_volume_container_2segments_square_ubottom(
        x=10, h_cuboid=10, liquid_height=15
      ),
      (2 / 3) * math.pi * (10 / 2) ** 3 + 10**2 * 10,
    )



from pylabrobot.resources.tip_rack import (
  NestedTipRack,
  TipRack,
  TipSpot,
)
from pylabrobot.resources.utils import create_ordered_items_2d

from .tip_creators import (
  hamilton_tip_10uL,
  hamilton_tip_10uL_filter,
  hamilton_tip_50uL,
  hamilton_tip_50uL_filter,
  hamilton_tip_300uL,
  hamilton_tip_300uL_filter,
  hamilton_tip_300uL_filter_slim,
  hamilton_tip_300uL_filter_ultrawide,
  hamilton_tip_1000uL,
  hamilton_tip_1000uL_filter,
  hamilton_tip_1000uL_filter_ultrawide,
  hamilton_tip_1000uL_filter_wide,
  hamilton_tip_4000uL_filter,
  hamilton_tip_5000uL,
)

# # # # # # # # # # 10 ul Tips # # # # # # # # # #


def hamilton_96_tiprack_10uL_filter(name: str, with_tips: bool = True) -> TipRack:
  """Hamilton cat. no.: 235936 (sterile), 235901 (non-sterile)
  Hamilton name: 'LTF'
  Tip Rack with 96x 10ul Low Volume Tip with filter
  """
  return TipRack(
    name=name,
    size_x=122.4,
    size_y=82.6,
    size_z=20.0,
    model=hamilton_96_tiprack_10uL_filter.__name__,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.2,
      dy=5.3,
      dz=-22.5,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=hamilton_tip_10uL_filter,
    ),
    with_tips=with_tips,
  )


# TODO: identify cat number
def hamilton_96_tiprack_10uL(name: str, with_tips: bool = True) -> TipRack:
  """Hamilton cat. no.: 235900 (non-sterile) 235935 (sterile)
  Hamilton name: 'LT'
  Tip Rack with 96x 10ul Low Volume Tip"""
  return TipRack(
    name=name,
    size_x=122.4,
    size_y=82.6,
    size_z=20.0,
    model=hamilton_96_tiprack_10uL.__name__,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.2,
      dy=5.3,
      dz=-22.5,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=hamilton_tip_10uL,
    ),
    with_tips=with_tips,
  )


# # # # # # # # # # 50 ul Tips # # # # # # # # # #


def hamilton_96_tiprack_50uL_filter(name: str, with_tips: bool = True) -> TipRack:
  """Hamilton cat. no.: 235948 (non-sterile), 235979 (sterile), 235829 (clear, non-sterile)
  Hamilton name: 'TIP_50ul_w_filter'
  Tip Rack with 96x 50ul Tip"""
  return TipRack(
    name=name,
    size_x=122.4,
    size_y=82.6,
    size_z=18.0,
    model=hamilton_96_tiprack_50uL_filter.__name__,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.2,
      dy=5.3,
      dz=-40.5,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=hamilton_tip_50uL_filter,
    ),
    with_tips=with_tips,
  )


def hamilton_96_tiprack_50uL(name: str, with_tips: bool = True) -> TipRack:
  """Hamilton cat. no.: 235966 (non-sterile) 235978 (sterile)
  Hamilton name: 'TIP_50ul'
  Tip Rack with 96x 50ul Tip no filter"""
  return TipRack(
    name=name,
    size_x=122.4,
    size_y=82.6,
    size_z=18.0,
    model=hamilton_96_tiprack_50uL.__name__,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.2,
      dy=5.3,
      dz=-40.5,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=hamilton_tip_50uL,
    ),
    with_tips=with_tips,
  )


def hamilton_96_tiprack_50uL_NTR(name: str, with_tips: bool = True) -> NestedTipRack:
  """Hamilton cat. no.: 235947 (non-sterile), 235964 (clear, non-sterile), 235987 (sterile)
  Nested Tip Rack with 96x 50ul Tips
  No filter
  """
  return NestedTipRack(
    name=name,
    size_x=127.76,
    size_y=85.48,
    size_z=56.0,  # Hamilton_96_tiprack_50ul_NTR + TIP_50ul_L.fitting_depth
    model=hamilton_96_tiprack_50uL_NTR.__name__,
    stacking_z_height=16.0,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=9.45,
      dy=7.55,
      dz=56.0 - 40.5 - 2,
      # top of Hamilton_96_tiprack_50ul_NTR - TIP_50ul_L.max_tip_length - "inbetween-space"(?)
      item_dx=9.0,
      item_dy=9.0,
      size_x=8.15,
      size_y=8.15,
      make_tip=hamilton_tip_50uL,
    ),
    with_tips=with_tips,
  )


# # # # # # # # # # 300 ul Tips # # # # # # # # # #


def hamilton_96_tiprack_300uL_filter(name: str, with_tips: bool = True) -> TipRack:
  """Hamilton cat. no.: 235830 (clear, non-sterile), 235903 (non-sterile), 235938 (sterile)
  Hamilton name: 'STF'
  Tip Rack with 96x 300ul Standard Volume Tip with filter"""
  return TipRack(
    name=name,
    size_x=122.4,
    size_y=82.6,
    size_z=20.0,
    model=hamilton_96_tiprack_300uL_filter.__name__,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.2,
      dy=5.3,
      dz=-50.5,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=hamilton_tip_300uL_filter,
    ),
    with_tips=with_tips,
  )


def hamilton_96_tiprack_300uL(name: str, with_tips: bool = True) -> TipRack:
  """Hamilton cat. no.: 235834 (clear, non-sterile), 235902 (non-sterile), 235937 (sterile)
  Hamilton name: 'ST'
  Tip Rack with 96x 300ul Standard Volume Tip"""
  return TipRack(
    name=name,
    size_x=122.4,
    size_y=82.6,
    size_z=20.0,
    model=hamilton_96_tiprack_300uL.__name__,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.2,
      dy=5.3,
      dz=-50.5,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=hamilton_tip_300uL,
    ),
    with_tips=with_tips,
  )


def hamilton_96_tiprack_300uL_filter_slim(name: str, with_tips: bool = True) -> TipRack:
  """Hamilton cat. no.: 235646 (CORE-II: conductive)
  Hamilton name: 'STF_Slim'
  Tip Rack with 96x 300ul Slim Standard Volume Tip with filter"""
  return TipRack(
    name=name,
    size_x=122.4,
    size_y=82.6,
    size_z=20.0,
    model=hamilton_96_tiprack_300uL_filter_slim.__name__,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.2,
      dy=5.3,
      dz=-83.5,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=hamilton_tip_300uL_filter_slim,
    ),
    with_tips=with_tips,
  )


def hamilton_96_tiprack_300uL_filter_ultrawide(name: str, with_tips: bool = True) -> TipRack:
  """Hamilton cat. no.: 235449 (1.55 mm oriface, non-sterile)
  Hamilton name: 'STF'
  Tip Rack with 96x 300ul Wide Bore Standard Volume Tip with filter"""
  return TipRack(
    name=name,
    size_x=122.4,
    size_y=82.6,
    size_z=20.0,
    model=hamilton_96_tiprack_300uL_filter_ultrawide.__name__,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.2,
      dy=5.3,
      dz=-42.5,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=hamilton_tip_300uL_filter_ultrawide,
    ),
    with_tips=with_tips,
  )


# # # # # # # # # # 1_000 uL Tips # # # # # # # # # #


def hamilton_96_tiprack_1000uL_filter(name: str, with_tips: bool = True) -> TipRack:
  """Hamilton cat. no.: 235820 (clear, non-sterile), 235905 (non-sterile), 235940 (sterile)
  Hamilton name: 'HTF'
  Tip Rack with 96x 1000ul High Volume Tip with filter
  """
  return TipRack(
    name=name,
    size_x=122.4,
    size_y=82.6,
    size_z=20.0,
    model=hamilton_96_tiprack_1000uL_filter.__name__,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.2,
      dy=5.3,
      dz=-83.5,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=hamilton_tip_1000uL_filter,
    ),
    with_tips=with_tips,
  )


def hamilton_96_tiprack_1000uL(name: str, with_tips: bool = True) -> TipRack:
  """Hamilton cat. no.: 235822 (clear, non-sterile), 235904 (non-sterile), 235939 (sterile)
  Hamilton name: 'HT'
  Tip Rack with 96x 1000ul High Volume Tip"""
  return TipRack(
    name=name,
    size_x=122.4,
    size_y=82.6,
    size_z=20.0,
    model=hamilton_96_tiprack_1000uL.__name__,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.2,
      dy=5.3,
      dz=-83.5,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=hamilton_tip_1000uL,
    ),
    with_tips=with_tips,
  )


def hamilton_96_tiprack_1000uL_filter_wide(name: str, with_tips: bool = True) -> TipRack:
  """Hamilton cat. no.:

  core-ii:
  - non-sterile, filter: 235678
  - sterile, filter: 235677
  core-i:
  - ?

  Hamilton name: 'HTF_WIDE'
  Tip Rack with 96x 1000ul High Volume Tip with filter

  Orifice Size: 1.2mm
  """

  return TipRack(
    name=name,
    size_x=122.4,
    size_y=82.6,
    size_z=20.0,
    model=hamilton_96_tiprack_1000uL_filter_wide.__name__,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.2,
      dy=5.3,
      dz=-80.35,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=hamilton_tip_1000uL_filter_wide,
    ),
    with_tips=with_tips,
  )


def hamilton_96_tiprack_1000uL_filter_ultrawide(name: str, with_tips: bool = True) -> TipRack:
  """Hamilton cat. no.:

  core-ii:
  - non-sterile, filter: 235541
  - sterile, filter: 235842
  core-i:
  - ?

  Hamilton name: 'HTF_ULTRAWIDE'
  Tip Rack with 96x 1000ul High Volume Tip with filter

  Orifice Size: 3.2mm
  """

  return TipRack(
    name=name,
    size_x=122.4,
    size_y=82.6,
    size_z=20.0,
    model=hamilton_96_tiprack_1000uL_filter_ultrawide.__name__,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=12,
      num_items_y=8,
      dx=7.2,
      dy=5.3,
      dz=-68.4,
      item_dx=9.0,
      item_dy=9.0,
      size_x=9.0,
      size_y=9.0,
      make_tip=hamilton_tip_1000uL_filter_ultrawide,
    ),
    with_tips=with_tips,
  )


# # # # # # # # # # 4 ml Tips # # # # # # # # # #


def hamilton_24_tiprack_4000uL_filter(name: str, with_tips: bool = True) -> TipRack:
  """Hamilton cat. no.: 184021 (non-sterile), 184023 (sterile)
  Hamilton name: 'FourmlTF'
  Tip Rack 24x 4ml Tip with Filter landscape oriented"""
  return TipRack(
    name=name,
    size_x=122.4,
    size_y=82.6,
    size_z=7.0,
    model=hamilton_24_tiprack_4000uL_filter.__name__,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=6,
      num_items_y=4,
      dx=7.3,
      dy=5.2,
      dz=-93.2,
      item_dx=18.0,
      item_dy=18.0,
      size_x=18.0,
      size_y=18.0,
      make_tip=hamilton_tip_4000uL_filter,
    ),
    with_tips=with_tips,
  )


# # # # # # # # # # 5 ml Tips # # # # # # # # # #


def hamilton_24_tiprack_5000uL(name: str, with_tips: bool = True) -> TipRack:
  """Hamilton cat. no.: 184020 (non-sterile), 184022 (sterile)
  Hamilton name: 'FivemlT'
  Tip Rack 24x 5ml Tip landscape oriented"""
  return TipRack(
    name=name,
    size_x=122.4,
    size_y=82.6,
    size_z=7.0,
    model=hamilton_24_tiprack_5000uL.__name__,
    ordered_items=create_ordered_items_2d(
      TipSpot,
      num_items_x=6,
      num_items_y=4,
      dx=7.3,
      dy=5.2,
      dz=-93.2,
      item_dx=18.0,
      item_dy=18.0,
      size_x=18.0,
      size_y=18.0,
      make_tip=hamilton_tip_5000uL,
    ),
    with_tips=with_tips,
  )


# # # # # # # # # # Deprecation Warnings # # # # # # # # # #

# TODO: remove after December 2025 (giving approx. 3 month transition period)


def LTF(name: str) -> TipRack:
  raise NotImplementedError("LTF is deprecated. use hamilton_96_tiprack_10uL_filter instead")


def LT(name: str) -> TipRack:
  raise NotImplementedError("LT is deprecated. use hamilton_96_tiprack_10uL instead")


def TIP_50ul_w_filter(name: str) -> TipRack:
  raise NotImplementedError(
    "TIP_50ul_w_filter is deprecated. use hamilton_96_tiprack_50uL_filter instead"
  )


def TIP_50ul(name: str) -> TipRack:
  raise NotImplementedError("TIP_50ul is deprecated. use hamilton_96_tiprack_50uL instead")


def STF(name: str) -> TipRack:
  raise NotImplementedError("STF is deprecated. use hamilton_96_tiprack_300uL_filter instead")


def ST(name: str) -> TipRack:
  raise NotImplementedError("ST is deprecated. use hamilton_96_tiprack_300uL instead")


def STF_Slim(name: str) -> TipRack:
  raise NotImplementedError(
    "STF_Slim is deprecated. use hamilton_96_tiprack_300uL_filter_slim instead"
  )


def HTF(name: str) -> TipRack:
  raise NotImplementedError("HTF is deprecated. use hamilton_96_tiprack_1000uL_filter instead")


def HT(name: str) -> TipRack:
  raise NotImplementedError("HT is deprecated. use hamilton_96_tiprack_1000uL instead")



import unittest
from unittest.mock import AsyncMock

from pylabrobot.liquid_handling import LiquidHandler
from pylabrobot.liquid_handling.backends.serializing_backend import (
  SerializingBackend,
)
from pylabrobot.resources import (
  PLT_CAR_L5AC_A00,
  TIP_CAR_480_A00,
  Coordinate,
  Cor_96_wellplate_360ul_Fb,
  STARLetDeck,
  hamilton_96_tiprack_300uL_filter,
  no_tip_tracking,
  no_volume_tracking,
)
from pylabrobot.serializer import serialize


class _TestSerializingBackend(SerializingBackend):
  send_command = AsyncMock()


class SerializingBackendTests(unittest.IsolatedAsyncioTestCase):
  """Tests for the serializing backend"""

  async def asyncSetUp(self) -> None:
    self.backend = _TestSerializingBackend(num_channels=8)
    self.backend.send_command.reset_mock()
    self.deck = STARLetDeck()
    self.lh = LiquidHandler(backend=self.backend, deck=self.deck)
    await self.lh.setup()

    self.tip_car = TIP_CAR_480_A00(name="tip carrier")
    self.tip_car[0] = self.tip_rack = hamilton_96_tiprack_300uL_filter(name="tip_rack_01")
    self.deck.assign_child_resource(self.tip_car, rails=1)

    self.plt_car = PLT_CAR_L5AC_A00(name="plate carrier")
    self.plt_car[0] = self.plate = Cor_96_wellplate_360ul_Fb(name="plate_01")
    self.plt_car[1] = self.other_plate = Cor_96_wellplate_360ul_Fb(name="plate_02")
    self.deck.assign_child_resource(self.plt_car, rails=9)

    self.backend.send_command.reset_mock()

    self.maxDiff = None

  async def test_pick_up_tips(self):
    tip_spot = self.tip_rack.get_item("A1")
    tip = tip_spot.get_tip()
    await self.lh.pick_up_tips([tip_spot])
    self.backend.send_command.assert_called_once_with(
      command="pick_up_tips",
      data={
        "channels": [
          {
            "resource_name": tip_spot.name,
            "offset": serialize(Coordinate.zero()),
            "tip": serialize(tip),
          }
        ],
        "use_channels": [0],
      },
    )

  async def test_drop_tips(self):
    tip_spot = self.tip_rack.get_item("A1")
    tip = tip_spot.get_tip()
    self.lh.update_head_state({0: tip})

    tips = self.tip_rack["A1"]
    with no_tip_tracking():
      await self.lh.drop_tips(tips)
    self.backend.send_command.assert_called_once_with(
      command="drop_tips",
      data={
        "channels": [
          {
            "resource_name": tip_spot.name,
            "offset": serialize(Coordinate.zero()),
            "tip": serialize(tip),
          }
        ],
        "use_channels": [0],
      },
    )

  async def test_aspirate(self):
    well = self.plate.get_item("A1")
    well.tracker.set_volume(10)
    tip = self.tip_rack.get_tip(0)
    self.lh.update_head_state({0: tip})
    self.backend.send_command.reset_mock()
    await self.lh.aspirate([well], vols=[10])
    self.backend.send_command.assert_called_once_with(
      command="aspirate",
      data={
        "channels": [
          {
            "resource_name": well.name,
            "offset": serialize(Coordinate.zero()),
            "tip": tip.serialize(),
            "volume": 10,
            "flow_rate": None,
            "liquid_height": None,
            "blow_out_air_volume": None,
            "mix": None,
          }
        ],
        "use_channels": [0],
      },
    )

  async def test_dispense(self):
    wells = self.plate["A1"]
    tip = self.tip_rack.get_tip(0)
    self.lh.update_head_state({0: tip})
    self.backend.send_command.reset_mock()
    with no_volume_tracking():
      await self.lh.dispense(wells, vols=[10])
    self.backend.send_command.assert_called_once_with(
      command="dispense",
      data={
        "channels": [
          {
            "resource_name": wells[0].name,
            "offset": serialize(Coordinate.zero()),
            "tip": tip.serialize(),
            "volume": 10,
            "flow_rate": None,
            "liquid_height": None,
            "blow_out_air_volume": None,
            "mix": None,
          }
        ],
        "use_channels": [0],
      },
    )

  async def test_pick_up_tips96(self):
    await self.lh.pick_up_tips96(self.tip_rack)
    self.backend.send_command.assert_called_once_with(
      command="pick_up_tips96",
      data={
        "resource_name": self.tip_rack.name,
        "offset": serialize(Coordinate.zero()),
      },
    )

  async def test_drop_tips96(self):
    await self.lh.pick_up_tips96(self.tip_rack)
    self.backend.send_command.reset_mock()

    await self.lh.drop_tips96(self.tip_rack)
    self.backend.send_command.assert_called_once_with(
      command="drop_tips96",
      data={
        "resource_name": self.tip_rack.name,
        "offset": serialize(Coordinate.zero()),
      },
    )

  async def test_aspirate96(self):
    await self.lh.pick_up_tips96(self.tip_rack)
    self.backend.send_command.reset_mock()

    tips = [channel.get_tip() for channel in self.lh.head96.values()]
    await self.lh.aspirate96(self.plate, volume=10)
    self.backend.send_command.assert_called_once_with(
      command="aspirate96",
      data={
        "aspiration": {
          "well_names": [well.name for well in self.plate.get_all_items()],
          "offset": serialize(Coordinate.zero()),
          "volume": 10,
          "flow_rate": None,
          "liquid_height": None,
          "blow_out_air_volume": None,
          "tips": [serialize(tip) for tip in tips],
        }
      },
    )

  async def test_dispense96(self):
    await self.lh.pick_up_tips96(self.tip_rack)
    tips = [channel.get_tip() for channel in self.lh.head96.values()]
    await self.lh.aspirate96(self.plate, volume=10)
    self.backend.send_command.reset_mock()

    await self.lh.dispense96(self.plate, volume=10)
    self.backend.send_command.assert_called_once_with(
      command="dispense96",
      data={
        "dispense": {
          "well_names": [well.name for well in self.plate.get_all_items()],
          "offset": serialize(Coordinate.zero()),
          "volume": 10.0,
          "flow_rate": None,
          "liquid_height": None,
          "blow_out_air_volume": None,
          "tips": [serialize(tip) for tip in tips],
        }
      },
    )

  async def test_move(self):
    to = Coordinate(600, 200, 200)
    await self.lh.move_plate(self.plate, to=to)

    # Should have called pick_up_resource and drop_resource
    calls = self.backend.send_command.call_args_list
    self.assertEqual(len(calls), 2)

    # Check pick_up_resource call
    self.assertEqual(calls[0].kwargs["command"], "pick_up_resource")
    self.assertEqual(
      calls[0].kwargs["data"],
      {
        "resource_name": self.plate.name,
        "offset": serialize(Coordinate.zero()),
        "pickup_distance_from_top": 9.87,
        "direction": "FRONT",
      },
    )

    # Check drop_resource call
    self.assertEqual(calls[1].kwargs["command"], "drop_resource")
    self.assertEqual(
      calls[1].kwargs["data"],
      {
        "resource_name": self.plate.name,
        "destination": serialize(to),
        "offset": serialize(Coordinate.zero()),
        "pickup_distance_from_top": 9.87,
        "pickup_direction": "FRONT",
        "drop_direction": "FRONT",
        "rotation": 0,
      },
    )



"""Backend that drives an Opentrons Thermocycler via the HTTP API."""

from typing import List, Optional, cast

from pylabrobot.thermocycling.backend import ThermocyclerBackend
from pylabrobot.thermocycling.standard import BlockStatus, LidStatus, Protocol

try:
  from ot_api.modules import (
    list_connected_modules,
    thermocycler_close_lid,
    thermocycler_deactivate_block,
    thermocycler_deactivate_lid,
    thermocycler_open_lid,
    thermocycler_run_profile_no_wait,
    thermocycler_set_block_temperature,
    thermocycler_set_lid_temperature,
  )

  USE_OT = True
except ImportError as e:
  USE_OT = False
  _OT_IMPORT_ERROR = e


class OpentronsThermocyclerBackend(ThermocyclerBackend):
  """HTTP-API backend for the Opentrons GEN-1/GEN-2 Thermocycler.

  All core functions are supported. run_profile() is fire-and-forget,
  since PCR runs can outlive the decorator's default timeout.
  """

  def __init__(self, opentrons_id: str):
    """Create a new backend bound to a specific thermocycler.

    Args:
      opentrons_id: The OT-API module "id" for your thermocycler.
    """
    super().__init__()  # Call parent constructor
    if not USE_OT:
      raise RuntimeError(
        "Opentrons is not installed. Please run pip install pylabrobot[opentrons]."
        f" Import error: {_OT_IMPORT_ERROR}."
      )
    self.opentrons_id = opentrons_id
    self._current_protocol: Optional[Protocol] = None

  async def setup(self):
    """No extra setup needed for HTTP-API thermocycler."""

  async def stop(self):
    """Gracefully deactivate both heaters."""
    await self.deactivate_block()
    await self.deactivate_lid()

  def serialize(self) -> dict:
    """Include the Opentrons module ID in serialized state."""
    return {**super().serialize(), "opentrons_id": self.opentrons_id}

  async def open_lid(self):
    """Open the thermocycler lid."""
    return thermocycler_open_lid(module_id=self.opentrons_id)

  async def close_lid(self):
    """Close the thermocycler lid."""
    return thermocycler_close_lid(module_id=self.opentrons_id)

  async def set_block_temperature(self, temperature: List[float]):
    """Set block temperature in °C. Only single unique temperature supported."""
    if len(set(temperature)) != 1:
      raise ValueError(
        f"Opentrons thermocycler only supports a single unique block temperature, got {set(temperature)}"
      )
    temp_value = temperature[0]
    return thermocycler_set_block_temperature(celsius=temp_value, module_id=self.opentrons_id)

  async def set_lid_temperature(self, temperature: List[float]):
    """Set lid temperature in °C. Only single unique temperature supported."""
    if len(set(temperature)) != 1:
      raise ValueError(
        f"Opentrons thermocycler only supports a single unique lid temperature, got {set(temperature)}"
      )
    temp_value = temperature[0]
    return thermocycler_set_lid_temperature(celsius=temp_value, module_id=self.opentrons_id)

  async def deactivate_block(self):
    """Deactivate the block heater."""
    return thermocycler_deactivate_block(module_id=self.opentrons_id)

  async def deactivate_lid(self):
    """Deactivate the lid heater."""
    return thermocycler_deactivate_lid(module_id=self.opentrons_id)

  async def run_protocol(self, protocol: Protocol, block_max_volume: float):
    """Enqueue and return immediately (no wait) the PCR profile command."""

    # flatten the protocol to a list of Steps
    # in opentrons, the "celsius" key is used instead of "temperature"
    # step.temperature is now a list, but Opentrons only supports single temperature
    ot_profile = []
    for stage in protocol.stages:
      for step in stage.steps:
        for _ in range(stage.repeats):
          if len(set(step.temperature)) != 1:
            raise ValueError(
              f"Opentrons thermocycler only supports a single unique temperature per step, got {set(step.temperature)}"
            )
          celsius = step.temperature[0]
          ot_profile.append({"celsius": celsius, "holdSeconds": step.hold_seconds})

    self._current_protocol = protocol

    return thermocycler_run_profile_no_wait(
      profile=ot_profile,
      block_max_volume=block_max_volume,
      module_id=self.opentrons_id,
    )

  def _find_module(self) -> dict:
    """Helper to locate this module's live-data dict."""
    for m in list_connected_modules():
      if m["id"] == self.opentrons_id:
        return cast(dict, m["data"])
    raise RuntimeError(f"Module '{self.opentrons_id}' not found")

  async def get_block_current_temperature(self) -> List[float]:
    return [cast(float, self._find_module()["currentTemperature"])]

  async def get_block_target_temperature(self) -> List[float]:
    target_temp = self._find_module().get("targetTemperature")
    if target_temp is None:
      raise RuntimeError("Block target temperature is not set. is a cycle running?")
    return [cast(float, target_temp)]

  async def get_lid_current_temperature(self) -> List[float]:
    return [cast(float, self._find_module()["lidTemperature"])]

  async def get_lid_target_temperature(self) -> List[float]:
    """Get the lid target temperature in °C. Raises RuntimeError if no target is active."""
    target_temp = self._find_module().get("lidTargetTemperature")
    if target_temp is None:
      raise RuntimeError("Lid target temperature is not set. is a cycle running?")
    return [cast(float, target_temp)]

  async def get_lid_open(self) -> bool:
    return cast(str, self._find_module()["lidStatus"]) == "open"

  async def get_lid_status(self) -> LidStatus:
    status = cast(str, self._find_module()["lidTemperatureStatus"])
    # Map Opentrons status strings to our enum
    if status == "holding at target":
      return LidStatus.HOLDING_AT_TARGET
    else:
      return LidStatus.IDLE

  async def get_block_status(self) -> BlockStatus:
    status = cast(str, self._find_module()["status"])
    # Map Opentrons status strings to our enum
    if status == "holding at target":
      return BlockStatus.HOLDING_AT_TARGET
    else:
      return BlockStatus.IDLE

  async def get_hold_time(self) -> float:
    return cast(float, self._find_module().get("holdTime", 0.0))

  async def get_current_cycle_index(self) -> int:
    """Get the zero-based index of the current cycle from the Opentrons API."""

    # https://github.com/PyLabRobot/pylabrobot/issues/632
    raise NotImplementedError('Opentrons "cycle" concept is not understood currently.')

    # Since we send a flattened list of steps, we have to recover the cycle index based
    # on the current step index and total step count.
    seen_steps = 0
    current_step = self.get_current_step_index()
    for stage in self._current_protocol.stages:
      for _ in stage.steps:
        if seen_steps == current_step:
          return  # TODO: what is a cycle in OT?
        seen_steps += 1

    raise RuntimeError("Current cycle index is not available. Is a profile running?")

  async def get_total_cycle_count(self) -> int:
    # https://github.com/PyLabRobot/pylabrobot/issues/632
    raise NotImplementedError('Opentrons "cycle" concept is not understood currently.')

  async def get_current_step_index(self) -> int:
    """Get the zero-based index of the current step from the Opentrons API."""
    # Opentrons API returns one-based, convert to zero-based
    step_index = self._find_module().get("currentStepIndex")
    if step_index is None:
      raise RuntimeError("Current step index is not available. Is a profile running?")
    return cast(int, step_index) - 1

  async def get_total_step_count(self) -> int:
    total_steps = self._find_module().get("totalStepCount")
    if total_steps is None:
      raise RuntimeError("Total step count is not available. Is a profile running?")
    return cast(int, total_steps)



from typing import Optional, Tuple

from pylabrobot.plate_reading.byonoy.byonoy_backend import ByonoyLuminescence96AutomateBackend
from pylabrobot.plate_reading.plate_reader import PlateReader
from pylabrobot.resources import Coordinate, PlateHolder, Resource, ResourceHolder
from pylabrobot.resources.barcode import Barcode
from pylabrobot.resources.rotation import Rotation


class _ByonoyLuminescenceReaderPlateHolder(PlateHolder):
  """Custom plate holder that checks if the reader sits on the parent base.
  This check is used to prevent crashes (moving plate onto holder while reader is on the base)."""

  def __init__(
    self,
    name: str,
    child_location: Coordinate = Coordinate.zero(),
    category: str = "plate_holder",
    model: Optional[str] = None,
  ):
    super().__init__(
      name=name,
      size_x=127.76,
      size_y=85.59,
      size_z=0,
      pedestal_size_z=0,
      child_location=child_location,
      category=category,
      model=model,
    )
    self._byonoy_base: Optional["ByonoyLuminescenceBaseUnit"] = None

  def check_can_drop_resource_here(self, resource: Resource, *, reassign: bool = True) -> None:
    if self._byonoy_base is None:
      raise RuntimeError(
        "Plate holder not assigned to a ByonoyLuminescenceBaseUnit. This should not happen."
      )

    if self._byonoy_base.reader_unit_holder.resource is not None:
      raise RuntimeError(
        f"Cannot drop resource {resource.name} onto plate holder while reader unit is on the base. "
        "Please remove the reader unit from the base before dropping a resource."
      )

    super().check_can_drop_resource_here(resource, reassign=reassign)


class ByonoyLuminescenceBaseUnit(Resource):
  """Base unit for the Byonoy L96/L96A luminescence reader.

  The base unit is a simple resource that holds a plate. The reader unit sits on top of it.
  """

  def __init__(
    self,
    name: str,
    size_x: float,
    size_y: float,
    size_z: float,
    plate_holder_child_location: Coordinate,
    reader_unit_holder_child_location: Coordinate,
    rotation: Optional[Rotation] = None,
    category: Optional[str] = None,
    model: Optional[str] = None,
    barcode: Optional[Barcode] = None,
  ):
    super().__init__(
      name=name,
      size_x=size_x,
      size_y=size_y,
      size_z=size_z,
      rotation=rotation,
      category=category,
      model=model,
      barcode=barcode,
    )

    self.plate_holder = _ByonoyLuminescenceReaderPlateHolder(
      name=self.name + "_plate_holder",
      child_location=plate_holder_child_location,
    )
    self.assign_child_resource(self.plate_holder, location=Coordinate.zero())

    self.reader_unit_holder = ResourceHolder(
      name=self.name + "_reader_unit_holder",
      size_x=size_x,
      size_y=size_y,
      size_z=0,
      child_location=reader_unit_holder_child_location,
    )
    self.assign_child_resource(self.reader_unit_holder, location=Coordinate.zero())

  def assign_child_resource(
    self, resource: Resource, location: Optional[Coordinate], reassign: bool = True
  ) -> None:
    if isinstance(resource, _ByonoyLuminescenceReaderPlateHolder):
      if self.plate_holder._byonoy_base is not None:
        raise ValueError("ByonoyBase can only have one plate holder assigned.")
      self.plate_holder._byonoy_base = self
    super().assign_child_resource(resource, location, reassign)

  def check_can_drop_resource_here(self, resource: Resource, *, reassign: bool = True) -> None:
    raise RuntimeError(
      "ByonoyBase does not support assigning child resources directly. "
      "Use the plate_holder or reader_unit_holder to assign plates and the reader unit, respectively."
    )


class ByonoyLuminescence96Automate(PlateReader):
  """Byonoy L96/L96A luminescence plate reader unit.

  This is the reader unit that sits on top of the base unit.
  """

  def __init__(
    self,
    name: str,
    size_x: float,
    size_y: float,
    size_z: float,
    preferred_pickup_location: Optional[Coordinate] = None,
  ):
    super().__init__(
      name=name,
      size_x=size_x,
      size_y=size_y,
      size_z=size_z,
      backend=ByonoyLuminescence96AutomateBackend(),
      model="Byonoy L96 Reader Unit",
      preferred_pickup_location=preferred_pickup_location,
    )


def byonoy_l96_reader_unit(name: str) -> ByonoyLuminescence96Automate:
  """Create a Byonoy L96 reader unit `PlateReader`.

  Note: L96 (non-automate) does not have a preferred pickup location.
  """
  return ByonoyLuminescence96Automate(
    name=name,
    size_x=139.7,  # caliper
    size_y=97.5,  # caliper
    size_z=35,  # force z probing
    preferred_pickup_location=None,
  )


def byonoy_l96_base_unit(name: str) -> ByonoyLuminescenceBaseUnit:
  """Create a Byonoy L96 base unit."""
  return ByonoyLuminescenceBaseUnit(
    name=name,
    size_x=139.7,  # caliper
    size_y=97.5,  # caliper
    size_z=9.4,  # force z probing
    plate_holder_child_location=Coordinate(x=6.25, y=6.1, z=2.64),  # caliper
    reader_unit_holder_child_location=Coordinate(x=0, y=0, z=7.2),  # z = 42.2 - 35
  )


def byonoy_l96(
  name: str, assign: bool = True
) -> Tuple[ByonoyLuminescenceBaseUnit, ByonoyLuminescence96Automate]:
  """Creates a ByonoyLuminescenceBaseUnit and a PlateReader instance for L96 (non-automate).

  Args:
    name: Base name for the resources.
    assign: If True, the reader unit is assigned to the base unit's reader_unit_holder.

  Returns:
    A tuple of (base_unit, reader_unit).
  """
  base_unit = byonoy_l96_base_unit(name=name + "_base")
  reader_unit = byonoy_l96_reader_unit(name=name + "_reader")
  if assign:
    base_unit.reader_unit_holder.assign_child_resource(reader_unit)
  return base_unit, reader_unit
